
#include "fp_utils.h"
#include <inttypes.h>
#include <sollya.h>

comp_binary_32_t power2_32(int64_t E) {
  comp_binary_32_t y;
  int64_t i;

  if (E < 0) return 1.0f / power2_32(-E);
  y = 1.0f;
  for (i=(int64_t) 0;i<E;i++) {
    y *= 2.0f;
  }
  return y;
}

comp_binary_64_t power2_64(int64_t E) {
  comp_binary_64_t y;
  int64_t i;

  if (E < 0) return 1.0 / power2_64(-E);
  y = 1.0;
  for (i=(int64_t) 0;i<E;i++) {
    y *= 2.0;
  }
  return y;
}

comp_binary_128_t power2_128(int64_t E) {
  comp_binary_128_t y;
  int64_t i;

  if (E < 0) return 1.0q / power2_128(-E);
  y = 1.0q;
  for (i=(int64_t) 0;i<E;i++) {
    y *= 2.0q;
  }
  return y;
}

comp_decimal_64_t power10_64(int64_t E) {
  comp_decimal_64_t y;
  int64_t i;

  if (E < 0) return 1.dd / power10_64(-E);
  y = 1.dd;
  for (i=(int64_t) 0;i<E;i++) {
    y *= 10.dd;
  }
  return y;
}

comp_decimal_128_t power10_128(int64_t E) {
  comp_decimal_128_t y;
  int64_t i;

  if (E < 0) return 1.dl / power10_128(-E);
  y = 1.dl;
  for (i=(int64_t) 0;i<E;i++) {
    y *= 10.dl;
  }
  return y;
}

void comp_string_binary_32(char * RESTRICT str,
			   int * RESTRICT sign_negative_out, int * RESTRICT class_out, int64_t * RESTRICT exponent_out, comp_uint32_t * RESTRICT significand_out,
			   int sign_negative_in, int64_t exponent_in, comp_uint32_t significand_in) {
  uint32_t m, X;
  comp_binary_32_t t1, t3, t4, t5, d;
  __comp_binary_32_caster_t x_cast;
  int64_t E1, E2;
  comp_fp_class_t class;

  m = comp_uint32_from_comp_uint32(significand_in);
  t1 = m;
  if (sign_negative_in) t1 = -t1;
  E1 = exponent_in >> 1;
  E2 = exponent_in - E1;
  t3 = power2_32(E1);
  t4 = power2_32(E2);
  t5 = t3 * t1;
  d = t4 * t5;
  x_cast.f = d;
  X = x_cast.i;
  sprintf(str, "0x%08" PRIx32, X);
  comp_fp_decompose_binary_32(sign_negative_out, &class, exponent_out, significand_out, d);
  *class_out = (int) class;
}

void comp_string_binary_64(char * RESTRICT str,
			   int * RESTRICT sign_negative_out, int * RESTRICT class_out, int64_t * RESTRICT exponent_out, comp_uint64_t * RESTRICT significand_out,
			   int sign_negative_in, int64_t exponent_in, comp_uint64_t significand_in) {
  uint64_t m, X;
  comp_binary_64_t t1, t3, t4, t5, d;
  __comp_binary_64_caster_t x_cast;
  int64_t E1, E2;
  comp_fp_class_t class;

  m = comp_uint64_from_comp_uint64(significand_in);
  t1 = m;
  if (sign_negative_in) t1 = -t1;
  E1 = exponent_in >> 1;
  E2 = exponent_in - E1;
  t3 = power2_64(E1);
  t4 = power2_64(E2);
  t5 = t3 * t1;
  d = t4 * t5;
  x_cast.f = d;
  X = x_cast.i;
  sprintf(str, "0x%016" PRIx64, X);
  comp_fp_decompose_binary_64(sign_negative_out, &class, exponent_out, significand_out, d);
  *class_out = (int) class;
}

void comp_string_binary_128(char * RESTRICT str,
			    int * RESTRICT sign_negative_out, int * RESTRICT class_out, int64_t * RESTRICT exponent_out, comp_uint128_t * RESTRICT significand_out,
			    int sign_negative_in, int64_t exponent_in, comp_uint128_t significand_in) {
  __uint128_t m, X;
  comp_binary_128_t t1, t3, t4, t5, d, t2;
  __comp_binary_128_caster_t x_cast;
  int64_t E1, E2;
  comp_fp_class_t class;
  uint64_t Xlo, Xhi;
  uint64_t mhi, mlo;

  m = comp_uint128_from_comp_uint128(significand_in);
  mlo = (uint64_t) m;
  mhi = (uint64_t) (m >> 64);
  t1 = mhi;
  t2 = mlo;
  t1 *= 18446744073709551616.0q;
  t1 += t2;
  if (sign_negative_in) t1 = -t1;
  E1 = exponent_in >> 1;
  E2 = exponent_in - E1;
  t3 = power2_128(E1);
  t4 = power2_128(E2);
  t5 = t3 * t1;
  d = t4 * t5;
  x_cast.f = d;
  X = x_cast.i;
  Xlo = (uint64_t) X;
  Xhi = (uint64_t) (X >> 64);
  sprintf(str, "0x%016" PRIx64 "%016" PRIx64, Xhi, Xlo);
  comp_fp_decompose_binary_128(sign_negative_out, &class, exponent_out, significand_out, d);
  *class_out = (int) class;
}

void comp_string_decimal_64(char * RESTRICT str,
			    int * RESTRICT sign_negative_out, int * RESTRICT class_out, int64_t * RESTRICT exponent_out, comp_uint64_t * RESTRICT significand_out,
			    int sign_negative_in, int64_t exponent_in, comp_uint64_t significand_in) {
  uint64_t m, X;
  comp_decimal_64_t t1, t3, t4, t5, d;
  __comp_decimal_64_caster_t x_cast;
  int64_t E1, E2;
  comp_fp_class_t class;

  m = comp_uint64_from_comp_uint64(significand_in);
  t1 = m;
  if (sign_negative_in) t1 = -t1;
  E1 = exponent_in >> 1;
  E2 = exponent_in - E1;
  t3 = power10_64(E1);
  t4 = power10_64(E2);
  t5 = t3 * t1;
  d = t4 * t5;
  x_cast.f = d;
  X = x_cast.i;
  sprintf(str, "0x%016" PRIx64, X);
  comp_fp_decompose_decimal_64(sign_negative_out, &class, exponent_out, significand_out, d);
  *class_out = (int) class;
}

void comp_string_decimal_128(char * RESTRICT str,
			     int * RESTRICT sign_negative_out, int * RESTRICT class_out, int64_t * RESTRICT exponent_out, comp_uint128_t * RESTRICT significand_out,
			     int sign_negative_in, int64_t exponent_in, comp_uint128_t significand_in) {
  __uint128_t m, X;
  comp_decimal_128_t t1, t3, t4, t5, d, t2;
  __comp_decimal_128_caster_t x_cast;
  int64_t E1, E2;
  comp_fp_class_t class;
  uint64_t Xlo, Xhi;
  uint64_t mhi, mlo;

  m = comp_uint128_from_comp_uint128(significand_in);
  mlo = (uint64_t) m;
  mhi = (uint64_t) (m >> 64);
  t1 = mhi;
  t2 = mlo;
  t1 *= 18446744073709551616.dl;
  t1 += t2;
  if (sign_negative_in) t1 = -t1;
  E1 = exponent_in >> 1;
  E2 = exponent_in - E1;
  t3 = power10_128(E1);
  t4 = power10_128(E2);
  t5 = t3 * t1;
  d = t4 * t5;
  x_cast.f = d;
  X = x_cast.i;
  Xlo = (uint64_t) X;
  Xhi = (uint64_t) (X >> 64);
  sprintf(str, "0x%016" PRIx64 "%016" PRIx64, Xhi, Xlo);
  comp_fp_decompose_decimal_128(sign_negative_out, &class, exponent_out, significand_out, d);
  *class_out = (int) class;
}

int string_binary_32_glue(sollya_obj_list_t *result, void **args) {
  char str[1024];
  int s, cl;
  int64_t E;
  comp_uint32_t Mout;
  uint32_t mantissa;
  uint32_t m;
  int tmp1;
  uint32_t tmp2;
  comp_uint32_t M;
  sollya_obj_list_t res;
  
  tmp1 = *((int *) (args[2]));
  tmp2 = tmp1;
  m = tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[3]));
  tmp2 = tmp1;
  m |= tmp2;
  M = comp_uint32_to_comp_uint32(m);
  comp_string_binary_32(str, &s, &cl, &E, &Mout, *((int *) (args[0])), (int64_t) *((int *) (args[1])), M);
  mantissa = comp_uint32_from_comp_uint32(Mout);
  res = NULL;
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(((uint64_t) mantissa)), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int64(E), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(s), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(cl), res);
  res = sollya_lib_construct_object_list(sollya_lib_string(str), res);
  *result = res;
  return 1;
}

int string_binary_64_glue(sollya_obj_list_t *result, void **args) {
  char str[1024];
  int s, cl;
  int64_t E;
  comp_uint64_t Mout;
  uint64_t mantissa;
  uint64_t m;
  int tmp1;
  uint64_t tmp2;
  comp_uint64_t M;
  sollya_obj_list_t res;
  
  tmp1 = *((int *) (args[2]));
  tmp2 = tmp1;
  m = tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[3]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[4]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[5]));
  tmp2 = tmp1;
  m |= tmp2;
  M = comp_uint64_to_comp_uint64(m);
  comp_string_binary_64(str, &s, &cl, &E, &Mout, *((int *) (args[0])), (int64_t) *((int *) (args[1])), M);
  mantissa = comp_uint64_from_comp_uint64(Mout);
  res = NULL;
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantissa), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int64(E), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(s), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(cl), res);
  res = sollya_lib_construct_object_list(sollya_lib_string(str), res);
  *result = res;
  return 1;
}

int string_binary_128_glue(sollya_obj_list_t *result, void **args) {
  char str[1024];
  int s, cl;
  int64_t E;
  comp_uint128_t Mout;
  __uint128_t mantissa;
  uint64_t mantHi, mantLo;
  __uint128_t m;
  int tmp1;
  __uint128_t tmp2;
  comp_uint128_t M;
  sollya_obj_list_t res;
  
  tmp1 = *((int *) (args[2]));
  tmp2 = tmp1;
  m = tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[3]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[4]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[5]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[6]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[7]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[8]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[9]));
  tmp2 = tmp1;
  m |= tmp2;
  M = comp_uint128_to_comp_uint128(m);
  comp_string_binary_128(str, &s, &cl, &E, &Mout, *((int *) (args[0])), (int64_t) *((int *) (args[1])), M);
  mantissa = comp_uint128_from_comp_uint128(Mout);
  mantLo = (uint64_t) mantissa;
  mantissa >>= 64;
  mantHi = (uint64_t) mantissa;
  res = NULL;
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantLo), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantHi), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int64(E), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(s), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(cl), res);
  res = sollya_lib_construct_object_list(sollya_lib_string(str), res);
  *result = res;
  return 1;
}

int string_decimal_64_glue(sollya_obj_list_t *result, void **args) {
  char str[1024];
  int s, cl;
  int64_t E;
  comp_uint64_t Mout;
  uint64_t mantissa;
  uint64_t m;
  int tmp1;
  uint64_t tmp2;
  comp_uint64_t M;
  sollya_obj_list_t res;
  
  tmp1 = *((int *) (args[2]));
  tmp2 = tmp1;
  m = tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[3]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[4]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[5]));
  tmp2 = tmp1;
  m |= tmp2;
  M = comp_uint64_to_comp_uint64(m);
  comp_string_decimal_64(str, &s, &cl, &E, &Mout, *((int *) (args[0])), (int64_t) *((int *) (args[1])), M);
  mantissa = comp_uint64_from_comp_uint64(Mout);
  res = NULL;
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantissa), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int64(E), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(s), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(cl), res);
  res = sollya_lib_construct_object_list(sollya_lib_string(str), res);
  *result = res;
  return 1;
}

int string_decimal_128_glue(sollya_obj_list_t *result, void **args) {
  char str[1024];
  int s, cl;
  int64_t E;
  comp_uint128_t Mout;
  __uint128_t mantissa;
  uint64_t mantHi, mantLo;
  __uint128_t m;
  int tmp1;
  __uint128_t tmp2;
  comp_uint128_t M;
  sollya_obj_list_t res;
  
  tmp1 = *((int *) (args[2]));
  tmp2 = tmp1;
  m = tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[3]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[4]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[5]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[6]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[7]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[8]));
  tmp2 = tmp1;
  m |= tmp2;
  m <<= 16;
  tmp1 = *((int *) (args[9]));
  tmp2 = tmp1;
  m |= tmp2;
  M = comp_uint128_to_comp_uint128(m);
  comp_string_decimal_128(str, &s, &cl, &E, &Mout, *((int *) (args[0])), (int64_t) *((int *) (args[1])), M);
  mantissa = comp_uint128_from_comp_uint128(Mout);
  mantLo = (uint64_t) mantissa;
  mantissa >>= 64;
  mantHi = (uint64_t) mantissa;
  res = NULL;
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantLo), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_uint64(mantHi), res); 
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int64(E), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(s), res);
  res = sollya_lib_construct_object_list(sollya_lib_constant_from_int(cl), res);
  res = sollya_lib_construct_object_list(sollya_lib_string(str), res);
  *result = res;
  return 1;
}

