#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <sched.h>

#include "fp_utils.h"

#define READ_BUFFER_SIZE   1024*16
#define CHUNK_SIZE         5000
#define TIMING_RUNS        70
#define TIMING_FACTOR      5.0
#define TIMING_START       10

/* time must be a pointer to an unsigned long long int */
#define READ_TIME_COUNTER(time)                              \
        __asm__ __volatile__(                                \
                "xorl %%eax,%%eax\n\t"                       \
                "cpuid\n\t"                                  \
                "rdtsc\n\t"                                  \
                "movl %%eax,(%0)\n\t"                        \
                "movl %%edx,4(%0)\n\t"                       \
                "xorl %%eax,%%eax\n\t"                       \
                "cpuid\n\t"                                  \
                : /* nothing */                              \
		: "S"((time))				     \
                : "eax", "ebx", "ecx", "edx", "memory")

enum __comp_argument_type_t_enum {
  COMP_BINARY32,
  COMP_BINARY64,
  COMP_BINARY128,
  COMP_DECIMAL64,
  COMP_DECIMAL128
};
typedef enum __comp_argument_type_t_enum comp_argument_type_t;

#define COMP_NUM_ARGUMENT_TYPES 5


union __comp_argument_value_t_union {
  comp_binary_32_t    b32;
  comp_binary_64_t    b64;
  comp_binary_128_t  b128;
  comp_decimal_64_t   d64;
  comp_decimal_128_t d128;
};
typedef union __comp_argument_value_t_union comp_argument_value_t;

struct __comp_argument_t_struct {
  comp_argument_type_t  type;
  comp_argument_value_t value;
};
typedef struct __comp_argument_t_struct comp_argument_t;

struct __comp_testpoint_t_struct {
  comp_argument_t arg1;
  comp_argument_t arg2;
  int expectedResult;
};
typedef struct __comp_testpoint_t_struct comp_testpoint_t;

struct __comp_chunk_t_struct {
  int number;
  comp_testpoint_t testPoints[CHUNK_SIZE];
};
typedef struct __comp_chunk_t_struct comp_chunk_t;

struct __comp_timing_point_t_struct {
  comp_argument_type_t arg1Type;
  comp_argument_type_t arg2Type;
  double min;
  double avg;
  double max;
  double sum;
  double minEmptyFunction;
  uint64_t testsRun;
  uint64_t emptyFunctionsRun;
};
typedef struct __comp_timing_point_t_struct comp_timing_point_t;

struct __comp_timing_data_t_struct {
  comp_timing_point_t timings[COMP_NUM_ARGUMENT_TYPES * COMP_NUM_ARGUMENT_TYPES];
  int timingRuns;
};
typedef struct __comp_timing_data_t_struct comp_timing_data_t;


extern int comp_le_bin32_bin32(comp_binary_32_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int comp_le_bin32_bin64(comp_binary_32_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int comp_le_bin32_bin128(comp_binary_32_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int comp_le_bin32_dec64(comp_binary_32_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int comp_le_bin32_dec128(comp_binary_32_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int comp_le_bin64_bin32(comp_binary_64_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int comp_le_bin64_bin64(comp_binary_64_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int comp_le_bin64_bin128(comp_binary_64_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int comp_le_bin64_dec64(comp_binary_64_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int comp_le_bin64_dec128(comp_binary_64_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int comp_le_bin128_bin32(comp_binary_128_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int comp_le_bin128_bin64(comp_binary_128_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int comp_le_bin128_bin128(comp_binary_128_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int comp_le_bin128_dec64(comp_binary_128_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int comp_le_bin128_dec128(comp_binary_128_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int comp_le_dec64_bin32(comp_decimal_64_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int comp_le_dec64_bin64(comp_decimal_64_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int comp_le_dec64_bin128(comp_decimal_64_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int comp_le_dec64_dec64(comp_decimal_64_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int comp_le_dec64_dec128(comp_decimal_64_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int comp_le_dec128_bin32(comp_decimal_128_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int comp_le_dec128_bin64(comp_decimal_128_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int comp_le_dec128_bin128(comp_decimal_128_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int comp_le_dec128_dec64(comp_decimal_128_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int comp_le_dec128_dec128(comp_decimal_128_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int empty_func_bin32_bin32(comp_binary_32_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int empty_func_bin32_bin64(comp_binary_32_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int empty_func_bin32_bin128(comp_binary_32_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int empty_func_bin32_dec64(comp_binary_32_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int empty_func_bin32_dec128(comp_binary_32_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int empty_func_bin64_bin32(comp_binary_64_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int empty_func_bin64_bin64(comp_binary_64_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int empty_func_bin64_bin128(comp_binary_64_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int empty_func_bin64_dec64(comp_binary_64_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int empty_func_bin64_dec128(comp_binary_64_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int empty_func_bin128_bin32(comp_binary_128_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int empty_func_bin128_bin64(comp_binary_128_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int empty_func_bin128_bin128(comp_binary_128_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int empty_func_bin128_dec64(comp_binary_128_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int empty_func_bin128_dec128(comp_binary_128_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int empty_func_dec64_bin32(comp_decimal_64_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int empty_func_dec64_bin64(comp_decimal_64_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int empty_func_dec64_bin128(comp_decimal_64_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int empty_func_dec64_dec64(comp_decimal_64_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int empty_func_dec64_dec128(comp_decimal_64_t a, comp_decimal_128_t b) __attribute__((weak)); 

extern int empty_func_dec128_bin32(comp_decimal_128_t a, comp_binary_32_t b) __attribute__((weak)); 
extern int empty_func_dec128_bin64(comp_decimal_128_t a, comp_binary_64_t b) __attribute__((weak)); 
extern int empty_func_dec128_bin128(comp_decimal_128_t a, comp_binary_128_t b) __attribute__((weak)); 
extern int empty_func_dec128_dec64(comp_decimal_128_t a, comp_decimal_64_t b) __attribute__((weak)); 
extern int empty_func_dec128_dec128(comp_decimal_128_t a, comp_decimal_128_t b) __attribute__((weak)); 



void *mycalloc(size_t nmemb, size_t size) {
  void *ptr;

  ptr = calloc(nmemb, size);
  if (ptr == NULL) {
    fprintf(stderr,"Could not allocate any additional memory; exiting\n");
    exit(17);
  } 

  return ptr;
}


char *readTestVectorLine(FILE *fd, char* buf, size_t bufSize, char **rest, size_t *restSize) {
  char *line = NULL;
  size_t lineSize = 0;
  size_t lineUsed = 0;
  char *curr;
  char *temp;
  size_t elementsRead;
  int ferr;
  size_t i;
  
  if (*restSize > 0) {
    line = (char *) mycalloc(*restSize + 1, sizeof(char));
    lineSize = *restSize + 1;
    lineUsed = 0;
    curr = line; 
    for (i=0;i<*restSize;i++) {
      if ((*rest)[i] == '\n') {
	*rest = &((*rest)[i + 1]);
	*restSize -= (i + 1);
	return line;
      }
      *curr = (*rest)[i];
      curr++;
      lineUsed++;
    }
  }

  while (1) {
    clearerr(fd);
    elementsRead = fread(buf, sizeof(char), bufSize, fd);
    ferr = ferror(fd);
    if (elementsRead != 0) {
      if (line == NULL) {
	line = (char *) mycalloc(elementsRead + 1, sizeof(char));
	lineSize = elementsRead + 1;
	lineUsed = 0;
	curr = line; 
      }
      for (i=0;i<elementsRead;i++) {
	if (buf[i] == '\n') {
	  *rest = &(buf[i + 1]);
	  *restSize = elementsRead - (i + 1);
	  return line;
	}
	if (lineUsed > lineSize - 2) {
	  temp = (char *) mycalloc(lineSize * 2 + 1, sizeof(char));
	  lineSize = lineSize * 2 + 1;
	  memcpy(temp, line, lineUsed);
	  free(line);
	  line = temp;
	  curr = line + lineUsed;
	}
	*curr = buf[i];
	curr++;
	lineUsed++;
      }
    } 
    if (elementsRead < bufSize) {
      if (feof(fd)) {
	break;
      } 
      /* Here, we could not read enough characters but we haven't reached eof */
      if (ferr != 0) {
	fprintf(stderr, "Some I/O error occurred while reading in the test vector file; exiting.\n");
	exit(17);
      }
      /* Hugh, no error occurred but we could read enough? Let's try again! */
    }
  }
  
  return line;
}

int parseArgumentType(comp_argument_t *arg, char *str) {
  if (strcmp(str, "b32") == 0) { 
    arg->type = COMP_BINARY32;
    return 1;
  }
  if (strcmp(str, "b64") == 0) { 
    arg->type = COMP_BINARY64;
    return 1;
  }
  if (strcmp(str, "b128") == 0) { 
    arg->type = COMP_BINARY128;
    return 1;
  }
  if (strcmp(str, "d64") == 0) { 
    arg->type = COMP_DECIMAL64;
    return 1;
  }
  if (strcmp(str, "d128") == 0) { 
    arg->type = COMP_DECIMAL128;
    return 1;
  }
  return 0;
}

int readHexadecimal(__uint128_t *val, char *str) {
  __uint128_t tmp;
  char *curr;
 
  if (str[0] != '0') return 0;
  if (str[1] != 'x') return 0;
  tmp = (__uint128_t) 0;
  for (curr=str+2; *curr != '\0'; curr++) {
    if ((*curr >= '0') && (*curr <= '9')) {
      tmp <<= 4;
      tmp += (__uint128_t) (*curr - '0');
    } else {
      if ((*curr >= 'A') && (*curr <= 'F')) {
	tmp <<= 4;
	tmp += (__uint128_t) (*curr - 'A' + 0xA);
      } else {
	if ((*curr >= 'a') && (*curr <= 'f')) {
	  tmp <<= 4;
	  tmp += (__uint128_t) (*curr - 'a' + 0xa);
	} else {
	  return 0;
	}
      }
    }
  }
  *val = tmp;
  return 1;
}

int parseArgument(comp_argument_t *arg, char *str) {
  __uint128_t tmp;
  uint32_t tmp32;
  uint64_t tmp64;
  __uint128_t tmp128;
  __comp_binary_32_caster_t tmpCasterBin32;
  __comp_binary_64_caster_t tmpCasterBin64;
  __comp_binary_128_caster_t tmpCasterBin128;
  __comp_decimal_64_caster_t tmpCasterDec64;
  __comp_decimal_128_caster_t tmpCasterDec128;
  
  if (!readHexadecimal(&tmp, str)) return 0;
  switch (arg->type) {
  case COMP_BINARY32:
    tmp32 = (uint32_t) tmp;
    tmpCasterBin32.i = tmp32;
    arg->value.b32 = tmpCasterBin32.f;
    break;
  case COMP_BINARY64:
    tmp64 = (uint64_t) tmp;
    tmpCasterBin64.i = tmp64;
    arg->value.b64 = tmpCasterBin64.f;
    break;
  case COMP_BINARY128:
    tmp128 = (__uint128_t) tmp;
    tmpCasterBin128.i = tmp128;
    arg->value.b128 = tmpCasterBin128.f;
    break;
  case COMP_DECIMAL64:
    tmp64 = (uint64_t) tmp;
    tmpCasterDec64.i = tmp64;
    arg->value.d64 = tmpCasterDec64.f;
    break;
  case COMP_DECIMAL128:
    tmp128 = (__uint128_t) tmp;
    tmpCasterDec128.i = tmp128;
    arg->value.d128 = tmpCasterDec128.f;
    break;
  default:
    return 0;
  }

  return 1;
}

int parseExpectedResult(int *arg, char *str) {
  if (strcmp(str, "true") == 0) {
    *arg = 1;
    return 1;
  }
  if (strcmp(str, "false") == 0) {
    *arg = 0;
    return 1;
  }  
  return 0;
}

void displayArgumentType(FILE *fd, comp_argument_type_t arg) {
  switch (arg) {
  case COMP_BINARY32:
    fprintf(fd, "b32");
    break;
  case COMP_BINARY64:
    fprintf(fd, "b64");
    break;
  case COMP_BINARY128:
    fprintf(fd, "b128");
    break;
  case COMP_DECIMAL64:
    fprintf(fd, "d64");
    break;
  case COMP_DECIMAL128:
    fprintf(fd, "d128");
    break;
  default:
    break;
  }
}

void displayArgumentValue(FILE *fd, comp_argument_t arg) {
  __comp_binary_32_caster_t tmpCasterBin32;
  __comp_binary_64_caster_t tmpCasterBin64;
  __comp_binary_128_caster_t tmpCasterBin128;
  __comp_decimal_64_caster_t tmpCasterDec64;
  __comp_decimal_128_caster_t tmpCasterDec128;
  __uint128_t tmp;
  uint64_t hi, lo;

  switch (arg.type) {
  case COMP_BINARY32:
    tmpCasterBin32.f = arg.value.b32;
    fprintf(fd, "0x%08" PRIx32, tmpCasterBin32.i);
    break;
  case COMP_BINARY64:
    tmpCasterBin64.f = arg.value.b64;
    fprintf(fd, "0x%016" PRIx64, tmpCasterBin64.i);
    break;
  case COMP_BINARY128:
    tmpCasterBin128.f = arg.value.b128;
    tmp = tmpCasterBin128.i;
    lo = (uint64_t) tmp;
    tmp >>= 64;
    hi = (uint64_t) tmp;
    fprintf(fd, "0x%016" PRIx64 "%016" PRIx64, hi, lo);
    break;
  case COMP_DECIMAL64:
    tmpCasterDec64.f = arg.value.d64;
    fprintf(fd, "0x%016" PRIx64, tmpCasterDec64.i);
    break;
  case COMP_DECIMAL128:
    tmpCasterDec128.f = arg.value.d128;
    tmp = tmpCasterDec128.i;
    lo = (uint64_t) tmp;
    tmp >>= 64;
    hi = (uint64_t) tmp;
    fprintf(fd, "0x%016" PRIx64 "%016" PRIx64, hi, lo);
    break;
  default:
    break;
  }
}

void displayTestCase(FILE *fd, comp_argument_t arg1, comp_argument_t arg2, int expectedResult) {
  displayArgumentType(fd, arg1.type);
  fprintf(fd, " ");
  displayArgumentType(fd, arg2.type);
  fprintf(fd, " ");
  displayArgumentValue(fd, arg1);
  fprintf(fd, " ");
  displayArgumentValue(fd, arg2);
  fprintf(fd, " ");
  if (expectedResult) fprintf(fd, "true"); else fprintf(fd, "false");
}

void displayWrongResult(comp_argument_t arg1, comp_argument_t arg2, int actualResult, int expectedResult) {
  printf("Wrong result for \"");
  displayTestCase(stdout, arg1, arg2, expectedResult);
  printf("\" - got ");
  if (actualResult) printf("true"); else printf("false");
  printf("\n");
}

int runFunction(comp_argument_t arg1, comp_argument_t arg2) {
  switch (arg1.type) {
  case COMP_BINARY32:
    switch (arg2.type) {
    case COMP_BINARY32:
      return comp_le_bin32_bin32(arg1.value.b32, arg2.value.b32);
      break;
    case COMP_BINARY64:
      return comp_le_bin32_bin64(arg1.value.b32, arg2.value.b64);
      break;
    case COMP_BINARY128:
      return comp_le_bin32_bin128(arg1.value.b32, arg2.value.b128);
      break;
    case COMP_DECIMAL64:
      return comp_le_bin32_dec64(arg1.value.b32, arg2.value.d64);
      break;
    case COMP_DECIMAL128:
      return comp_le_bin32_dec128(arg1.value.b32, arg2.value.d128);
      break;
    default:
      return 0;
    }
    break;
  case COMP_BINARY64:
    switch (arg2.type) {
    case COMP_BINARY32:
      return comp_le_bin64_bin32(arg1.value.b64, arg2.value.b32);
      break;
    case COMP_BINARY64:
      return comp_le_bin64_bin64(arg1.value.b64, arg2.value.b64);
      break;
    case COMP_BINARY128:
      return comp_le_bin64_bin128(arg1.value.b64, arg2.value.b128);
      break;
    case COMP_DECIMAL64:
      return comp_le_bin64_dec64(arg1.value.b64, arg2.value.d64);
      break;
    case COMP_DECIMAL128:
      return comp_le_bin64_dec128(arg1.value.b64, arg2.value.d128);
      break;
    default:
      return 0;
    }
    break;
  case COMP_BINARY128:
    switch (arg2.type) {
    case COMP_BINARY32:
      return comp_le_bin128_bin32(arg1.value.b128, arg2.value.b32);
      break;
    case COMP_BINARY64:
      return comp_le_bin128_bin64(arg1.value.b128, arg2.value.b64);
      break;
    case COMP_BINARY128:
      return comp_le_bin128_bin128(arg1.value.b128, arg2.value.b128);
      break;
    case COMP_DECIMAL64:
      return comp_le_bin128_dec64(arg1.value.b128, arg2.value.d64);
      break;
    case COMP_DECIMAL128:
      return comp_le_bin128_dec128(arg1.value.b128, arg2.value.d128);
      break;
    default:
      return 0;
    }
    break;
  case COMP_DECIMAL64:
    switch (arg2.type) {
    case COMP_BINARY32:
      return comp_le_dec64_bin32(arg1.value.d64, arg2.value.b32);
      break;
    case COMP_BINARY64:
      return comp_le_dec64_bin64(arg1.value.d64, arg2.value.b64);
      break;
    case COMP_BINARY128:
      return comp_le_dec64_bin128(arg1.value.d64, arg2.value.b128);
      break;
    case COMP_DECIMAL64:
      return comp_le_dec64_dec64(arg1.value.d64, arg2.value.d64);
      break;
    case COMP_DECIMAL128:
      return comp_le_dec64_dec128(arg1.value.d64, arg2.value.d128);
      break;
    default:
      return 0;
    }
    break;
  case COMP_DECIMAL128:
    switch (arg2.type) {
    case COMP_BINARY32:
      return comp_le_dec128_bin32(arg1.value.d128, arg2.value.b32);
      break;
    case COMP_BINARY64:
      return comp_le_dec128_bin64(arg1.value.d128, arg2.value.b64);
      break;
    case COMP_BINARY128:
      return comp_le_dec128_bin128(arg1.value.d128, arg2.value.b128);
      break;
    case COMP_DECIMAL64:
      return comp_le_dec128_dec64(arg1.value.d128, arg2.value.d64);
      break;
    case COMP_DECIMAL128:
      return comp_le_dec128_dec128(arg1.value.d128, arg2.value.d128);
      break;
    default:
      return 0;
    }
    break;
  default:
    return 0;
  }
  return 0;
}

void integrateTimingPoint(comp_timing_point_t *timingPoint, uint64_t cycles, comp_argument_t arg1, comp_argument_t arg2) {
  double cyclesAsDouble, avg;

  if (timingPoint == NULL) return;
  if (timingPoint->arg1Type != arg1.type) return;
  if (timingPoint->arg2Type != arg2.type) return;
  cyclesAsDouble = (double) cycles;
  if (timingPoint->testsRun > TIMING_START) {
    avg = timingPoint->sum / ((double) timingPoint->testsRun) - timingPoint->minEmptyFunction;
    if ((cyclesAsDouble - timingPoint->minEmptyFunction) < (avg / TIMING_FACTOR)) return;
    if ((cyclesAsDouble - timingPoint->minEmptyFunction) > (avg * TIMING_FACTOR)) return;
    if (cyclesAsDouble < timingPoint->min) timingPoint->min = cyclesAsDouble;
    if (cyclesAsDouble > timingPoint->max) timingPoint->max = cyclesAsDouble;
  }
  timingPoint->sum += cyclesAsDouble;
  (timingPoint->testsRun)++;
}

void integrateEmptyFunctionTimingPoint(comp_timing_point_t *timingPoint, uint64_t cycles, comp_argument_t arg1, comp_argument_t arg2) {
  double cyclesAsDouble;

  if (timingPoint == NULL) return;
  if (timingPoint->arg1Type != arg1.type) return;
  if (timingPoint->arg2Type != arg2.type) return;
  cyclesAsDouble = (double) cycles;
  if (timingPoint->emptyFunctionsRun > TIMING_START) {
    if (cyclesAsDouble < timingPoint->minEmptyFunction) timingPoint->minEmptyFunction = cyclesAsDouble;
  }
  (timingPoint->emptyFunctionsRun)++;
}

void integrateTimingResults(comp_timing_data_t *timings, uint64_t cycles, comp_argument_t arg1, comp_argument_t arg2) {
  if (timings == NULL) return;
  integrateTimingPoint(&(timings->timings[COMP_NUM_ARGUMENT_TYPES * arg1.type + arg2.type]),cycles,arg1,arg2);
}

void integrateEmptyFunctionTimingResults(comp_timing_data_t *timings, uint64_t cycles, comp_argument_t arg1, comp_argument_t arg2) {
  if (timings == NULL) return;
  integrateEmptyFunctionTimingPoint(&(timings->timings[COMP_NUM_ARGUMENT_TYPES * arg1.type + arg2.type]),cycles,arg1,arg2);
}

void timeFunction(comp_timing_data_t *timings, comp_argument_t arg1, comp_argument_t arg2) {
  int res;
  volatile int sink;
  unsigned long long int before, after;
  uint64_t cycles;

  switch (arg1.type) {
  case COMP_BINARY32:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin32_bin32(arg1.value.b32, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin32_bin64(arg1.value.b32, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin32_bin128(arg1.value.b32, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin32_dec64(arg1.value.b32, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin32_dec128(arg1.value.b32, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_BINARY64:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin64_bin32(arg1.value.b64, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin64_bin64(arg1.value.b64, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin64_bin128(arg1.value.b64, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin64_dec64(arg1.value.b64, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin64_dec128(arg1.value.b64, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_BINARY128:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin128_bin32(arg1.value.b128, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin128_bin64(arg1.value.b128, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin128_bin128(arg1.value.b128, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin128_dec64(arg1.value.b128, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = comp_le_bin128_dec128(arg1.value.b128, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_DECIMAL64:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec64_bin32(arg1.value.d64, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec64_bin64(arg1.value.d64, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec64_bin128(arg1.value.d64, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec64_dec64(arg1.value.d64, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec64_dec128(arg1.value.d64, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_DECIMAL128:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec128_bin32(arg1.value.d128, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec128_bin64(arg1.value.d128, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec128_bin128(arg1.value.d128, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec128_dec64(arg1.value.d128, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = comp_le_dec128_dec128(arg1.value.d128, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  default:
    return;
  }
  /* "Use" the result */
  sink = res;
  (void) sink;

  /* Compute the cycle count */
  if (after < before) return;
  cycles = (uint64_t) (after - before);
  
  /* Integrate the cycle count into the known timings */
  integrateTimingResults(timings, cycles, arg1, arg2);
}

void timeEmptyFunction(comp_timing_data_t *timings, comp_argument_t arg1, comp_argument_t arg2) {
  int res;
  volatile int sink;
  unsigned long long int before, after;
  uint64_t cycles;

  switch (arg1.type) {
  case COMP_BINARY32:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin32_bin32(arg1.value.b32, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin32_bin64(arg1.value.b32, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin32_bin128(arg1.value.b32, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin32_dec64(arg1.value.b32, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin32_dec128(arg1.value.b32, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_BINARY64:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin64_bin32(arg1.value.b64, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin64_bin64(arg1.value.b64, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin64_bin128(arg1.value.b64, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin64_dec64(arg1.value.b64, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin64_dec128(arg1.value.b64, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_BINARY128:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin128_bin32(arg1.value.b128, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin128_bin64(arg1.value.b128, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin128_bin128(arg1.value.b128, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin128_dec64(arg1.value.b128, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = empty_func_bin128_dec128(arg1.value.b128, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_DECIMAL64:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec64_bin32(arg1.value.d64, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec64_bin64(arg1.value.d64, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec64_bin128(arg1.value.d64, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec64_dec64(arg1.value.d64, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec64_dec128(arg1.value.d64, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  case COMP_DECIMAL128:
    switch (arg2.type) {
    case COMP_BINARY32:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec128_bin32(arg1.value.d128, arg2.value.b32);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY64:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec128_bin64(arg1.value.d128, arg2.value.b64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_BINARY128:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec128_bin128(arg1.value.d128, arg2.value.b128);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL64:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec128_dec64(arg1.value.d128, arg2.value.d64);
      READ_TIME_COUNTER(&after);
      break;
    case COMP_DECIMAL128:
      READ_TIME_COUNTER(&before);
      res = empty_func_dec128_dec128(arg1.value.d128, arg2.value.d128);
      READ_TIME_COUNTER(&after);
      break;
    default:
      return;
    }
    break;
  default:
    return;
  }
  /* "Use" the result */
  sink = res;
  (void) sink;

  /* Compute the cycle count */
  if (after < before) return;
  cycles = (uint64_t) (after - before);
  
  /* Integrate the cycle count into the known timings */
  integrateEmptyFunctionTimingResults(timings, cycles, arg1, arg2);
}

int runTest(comp_argument_t arg1, comp_argument_t arg2, int expectedResult, int *numAll, int *numErr) {
  int actualResult;
  int result;

  (*numAll)++;
  actualResult = runFunction(arg1, arg2);
  result = ((!!actualResult) == (!!expectedResult));

  if (!result) {
    (*numErr)++;
    displayWrongResult(arg1, arg2, actualResult, expectedResult);
  }

  return result;
}

void doTiming(comp_timing_data_t *timings, comp_chunk_t *chunk) {
  int j, i;

  for (j=0;j<timings->timingRuns;j++) {
    for (i=0;i<chunk->number;i++) {
      timeEmptyFunction(NULL, chunk->testPoints[i].arg1, chunk->testPoints[i].arg2);
    }
  }
  for (j=0;j<timings->timingRuns;j++) {
    for (i=0;i<chunk->number;i++) {
      timeEmptyFunction(timings, chunk->testPoints[i].arg1, chunk->testPoints[i].arg2);
    }
  }
  for (j=0;j<timings->timingRuns;j++) {
    for (i=0;i<chunk->number;i++) {
      timeFunction(NULL, chunk->testPoints[i].arg1, chunk->testPoints[i].arg2);
    }
  }
  for (j=0;j<timings->timingRuns;j++) {
    for (i=0;i<chunk->number;i++) {
      timeFunction(timings, chunk->testPoints[i].arg1, chunk->testPoints[i].arg2);
    }
  }
}

void initializeTimingPoint(comp_timing_point_t *timingPoint, comp_argument_type_t arg1Type, comp_argument_type_t arg2Type) {
  timingPoint->arg1Type = arg1Type;
  timingPoint->arg2Type = arg2Type;
  timingPoint->min = 1e300 * 1e300;
  timingPoint->avg = 0.0;
  timingPoint->max = 0.0;
  timingPoint->sum = 0.0;
  timingPoint->minEmptyFunction = 1e300 * 1e300;
  timingPoint->testsRun = (uint64_t) 0;
  timingPoint->emptyFunctionsRun = (uint64_t) 0;
}

double maxDouble(double a, double b) {
  if (a > b) return a;
  return b;
}

void displayTimingPoint(comp_timing_point_t *timingPoint) {
  if (timingPoint->testsRun == ((uint64_t) 0)) return;
  if (timingPoint->emptyFunctionsRun == ((uint64_t) 0)) return;
  timingPoint->avg = timingPoint->sum / ((double) timingPoint->testsRun);
  printf("Timing result for case ");
  displayArgumentType(stdout, timingPoint->arg1Type);
  printf("/");
  displayArgumentType(stdout, timingPoint->arg2Type);
  printf(":\n");
  printf("func call time %8.3f\n",timingPoint->minEmptyFunction);
  printf("min            %8.3f\n",maxDouble(timingPoint->min - timingPoint->minEmptyFunction, 0.0));
  printf("avg            %8.3f\n",maxDouble(timingPoint->avg - timingPoint->minEmptyFunction, 0.0));  
  printf("max            %8.3f\n",maxDouble(timingPoint->max - timingPoint->minEmptyFunction, 0.0));  
  printf("\n");
}

void initializeTimings(comp_timing_data_t *timings, int timingRuns) {
  comp_argument_type_t i, j;
  
  for (i=COMP_BINARY32;i<=COMP_DECIMAL128;i++) {
    for (j=COMP_BINARY32;j<=COMP_DECIMAL128;j++) {
      initializeTimingPoint(&(timings->timings[COMP_NUM_ARGUMENT_TYPES * i + j]),i,j);
    }
  }
  timings->timingRuns = timingRuns;
}

void displayTimings(comp_timing_data_t *timings) {
  comp_argument_type_t i, j;
  
  for (i=COMP_BINARY32;i<=COMP_DECIMAL128;i++) {
    for (j=COMP_BINARY32;j<=COMP_DECIMAL128;j++) {
      displayTimingPoint(&(timings->timings[COMP_NUM_ARGUMENT_TYPES * i + j]));
    }
  }  
}

void runChunk(int *res, comp_chunk_t *chunk, comp_timing_data_t *timings, int *numAll, int *numErr, int displayProgress) {
  int i, testResult;

  for (i=0;i<chunk->number;i++) {
    testResult = runTest(chunk->testPoints[i].arg1, chunk->testPoints[i].arg2, chunk->testPoints[i].expectedResult, numAll, numErr);
    *res = *res && testResult;
  }
  if (displayProgress) {
    fprintf(stderr,"Running test %16d, %16d errors so far (\"", *numAll, *numErr);
    displayTestCase(stderr, 
		    chunk->testPoints[chunk->number-1].arg1, 
		    chunk->testPoints[chunk->number-1].arg2, 
		    chunk->testPoints[chunk->number-1].expectedResult);
    fprintf(stderr,"\")\n");
  }
  
  

  if (timings != NULL) {
    doTiming(timings, chunk);
  }
  chunk->number = 0;
}

void runTestOnChunk(int *res, comp_chunk_t *chunk, comp_argument_t arg1, comp_argument_t arg2, int expectedResult, comp_timing_data_t *timings, int *numAll, int *numErr, int displayProgress) {
  comp_testpoint_t testpoint;

  testpoint.arg1 = arg1;
  testpoint.arg2 = arg2;
  testpoint.expectedResult = expectedResult;
  if (chunk->number >= CHUNK_SIZE) {
    runChunk(res, chunk, timings, numAll, numErr, displayProgress);
  }
  chunk->testPoints[chunk->number] = testpoint;
  chunk->number++;
}

int runWords(int *res, comp_chunk_t *chunk, char **words, int numWords, comp_timing_data_t *timings, int *numAll, int *numErr, int displayProgress) {
  comp_argument_t arg1, arg2;
  int expectedResult;

  if (numWords < 1) return 0;
  if (strcmp(words[0], "//") == 0) return 1;
  if (numWords < 5) return 0;
  if (!parseArgumentType(&arg1, words[0])) return 0;
  if (!parseArgumentType(&arg2, words[1])) return 0;
  if (!parseArgument(&arg1, words[2])) return 0;
  if (!parseArgument(&arg2, words[3])) return 0;
  if (!parseExpectedResult(&expectedResult, words[4])) return 0;
  runTestOnChunk(res, chunk, arg1, arg2, expectedResult, timings, numAll, numErr, displayProgress);

  return 1;
}

int runLine(int *res, comp_chunk_t *chunk, char *line, comp_timing_data_t *timings, int *numAll, int *numErr, int displayProgress) {
  size_t len;
  char *myLine;
  char **words;
  int numWords;
  char *ptr;
  int state;
  int result;

  len = strlen(line);
  if ((len >= 2) &&
      (line[0] == '/') && 
      (line[1] == '/')) return 1;
  myLine = mycalloc(len + 1, sizeof(char));
  strcpy(myLine, line);
  words = mycalloc(len, sizeof(char**));
  numWords = 0;
  state = 0; /* Looking for start of word */
  for (ptr = myLine; *ptr != '\0'; ptr++) {
    switch (state) {
    case 0: /* Looking for start of word */
      if (*ptr != ' ') {
	/* Found a non-blank */
	numWords++;
	words[numWords - 1] = ptr;
	state = 1;
      }
      break;
    case 1: /* Looking for start of blanks */
      if (*ptr == ' ') {
	/* Found a blank */
	*ptr = '\0';
	state = 0;
      }
      break;
    }
  }
  result = runWords(res, chunk, words, numWords, timings, numAll, numErr, displayProgress);
  free(myLine);
  free(words);
  return result;
}

int readAndTest(FILE *fd, comp_timing_data_t *timings, int *numAll, int *numErr, int displayProgress) {
  char *currLine;
  char readBuffer[READ_BUFFER_SIZE];
  char *bufRest = NULL;
  size_t bufRestSize = 0;
  int res = 1;
  comp_chunk_t chunk;

  chunk.number = 0;
  while (1) {
    /* Read a line */
    if ((currLine = readTestVectorLine(fd, readBuffer, READ_BUFFER_SIZE, &bufRest, &bufRestSize)) == NULL) {
      /* Could not read a line anymore */
      runChunk(&res, &chunk, timings, numAll, numErr, displayProgress);
      return res;
    }
    /* Parse it */
    if (!runLine(&res, &chunk, currLine, timings, numAll, numErr, displayProgress)) {
      /* Could not parse the line, skip it */
      fprintf(stderr, "Could not parse the test vector line \"%s\", skipping it.\n", currLine);
    } 
    /* Free the line */
    free(currLine);
  }

  if (bufRestSize > 0) {
    if (fseek(fd, -((long) bufRestSize), SEEK_CUR) != 0) {
      fprintf(stderr, "Error during fseek: %s\n", strerror(errno));
    } 
  }

  runChunk(&res, &chunk, timings, numAll, numErr, displayProgress);
  return res;
}

int main(int argc, char **argv) {
  FILE *fd;
  int res, numAll, numErr, displayProgress;
  comp_timing_data_t *timings;
  comp_timing_data_t timingData;

  fd = stdin;
  displayProgress = 0;
  timings = NULL;
  if (argc >= 2) {
    if (strcmp(argv[1], "--progress") == 0) {
      displayProgress = 1;
    }
    if (strcmp(argv[1], "--time") == 0) {
      timings = &timingData;
    }    
  }
  if (argc >= 3) {
    if (strcmp(argv[2], "--progress") == 0) {
      displayProgress = 1;
    }
    if (strcmp(argv[2], "--time") == 0) {
      timings = &timingData;
    }    
  }

  if (timings != NULL) {
    initializeTimings(timings, TIMING_RUNS);
        cpu_set_t cpuMask;
        CPU_ZERO(&cpuMask);
        CPU_SET(1, &cpuMask);
        sched_setaffinity(0, sizeof(cpuMask), &cpuMask);
  }
  numAll = 0; numErr = 0;
  res = readAndTest(fd, timings, &numAll, &numErr, displayProgress);

  if (res) {
    printf("All %d tests have yielded the correct result.\n", numAll);
  } else {
    printf("There have been %d incorrect results, out of %d tests run.\n", numErr, numAll);
  }
  if (timings != NULL) {
    displayTimings(timings);
  }

  if (!res) return 1;
  return 0;
}
